@@ -4,7 +4,7 @@ use warnings; use strict; use utf8;
 #use 5.010; # Needed for smart-match operator
 #v5.10 feature use removed until 5.10 is standard on MacOSX and Debian
 use English qw( -no_match_vars );
-use version 0.74; our $VERSION = qv("0.4.7");
+use version 0.74; our $VERSION = qv("0.5.0");
 
 # Perl Critic overrides:
 ## no critic (Package variable)
@@ -494,13 +494,12 @@ sub write_text :method
     return unless($$self{text});
 
     debug(1,"DEBUG: writing text to '",$filename,
-          "', encoding ",$pdbencoding{$$self{encoding}});
+          "', encoding ",$pdbencoding{$self->{encoding}});
 
     open(my $fh,">",$filename)
         or croak($subname,"(): unable to open '",$filename,"' for writing!\n");
-    if($$self{encoding} == 65001) { binmode($fh,":utf8"); }
-    else { binmode($fh); }
-    print {*$fh} $$self{text};
+    binmode($fh);
+    print {*$fh} $self->{text};
     close($fh)
         or croak($subname,"(): unable to close '",$filename,"'!\n");
 
@@ -661,7 +660,7 @@ sub ParseRecord :method   ## no critic (Always unpack @_ first)
         else
         {
             debug(3,"DEBUG: record ",$currentrecord,
-                  " is PalmDoc-compressed text");
+                  " is uncompressed or PalmDoc-compressed text");
             $self->ParseRecordText(\$data);
         }
     }
@@ -1220,23 +1219,39 @@ sub ParseRecordText :method
     my $currentrecord = scalar @{$$self{records}};
     my $compression = $$self{header}{palm}{compression};
     my $recordtext;
+    my $extradatasize;
+
+    debug(3,"DEBUG: extradataflags = ",$$self{header}{mobi}{extradataflags});
+    if( $$self{header}{mobi}{extradataflags} )
+    {
+        $extradatasize = record_extradata_size(
+            dataref => $dataref,
+            extradataflags => $$self{header}{mobi}{extradataflags}
+           );
+    }
 
     if($compression == 1)        # No compression
     {
-        $recordtext = $$dataref;
+        debug(3,"DEBUG: No compression on record ",$currentrecord);
+        if($extradatasize)
+        {
+            $recordtext = substr($$dataref,0,length($$dataref)-$extradatasize);
+            debug(3,"DEBUG: skipping ",$extradatasize," bytes at end of record ",
+                  $currentrecord);
+        }
+        else
+        {
+            $recordtext = $$dataref;
+        }
     }
     elsif($compression == 2)     # PalmDoc compression
     {
         my %args;
-        my $extradatasize;
-
-        if( $$self{header}{mobi}{extradataflags} )
+        if($extradatasize)
         {
-            $extradatasize = record_extradata_size(
-                dataref => $dataref,
-                extradataflags => $$self{header}{mobi}{extradataflags}
-               );
             $args{trailing} = $extradatasize;
+            debug(3,"DEBUG: skipping ",$extradatasize," bytes at end of record ",
+                  $currentrecord);
         }
         $recordtext = uncompress_palmdoc($$dataref,%args);
     }
@@ -1257,12 +1272,12 @@ sub ParseRecordText :method
     if($recordtext)
     {
         $self->{text} .= $recordtext;
-        debug(3,"DEBUG: uncompressed text record ",$currentrecord);
+        debug(3,"DEBUG: parsed text record ",$currentrecord);
     }
     else
     {
         debug(1,"DEBUG: record ",$currentrecord,
-              " could not be decompressed (",
+              " could not be parsed (",
               length($$dataref)," bytes starting with '",
               substr($$dataref,0,4),"')");
         $$self{unknowndata}{$currentrecord} = $$dataref;
@@ -1321,6 +1336,8 @@ sub fix_html :method   ## no critic (Always unpack @_ first)
 
     my $tree;
     my @elements;
+    my $head_element;
+    my $element;
     my $recindex;
     my $link;
 
@@ -1331,31 +1348,59 @@ sub fix_html :method   ## no critic (Always unpack @_ first)
     $self->fix_html_filepos();
 
     # Convert or remove the Mobipocket-specific tags
-    $$self{text} =~ s#<mbp:pagebreak [\s\n]*
-                     #<br style="page-break-after: always" #gix;
-    $$self{text} =~ s#</mbp:pagebreak>##gix;
-    $$self{text} =~ s#</?mbp:nu>##gix;
-    $$self{text} =~ s#</?mbp:section>##gix;
-    $$self{text} =~ s#</?mbp:frameset##gix;
-    $$self{text} =~ s#</?mbp:slave-frame##gix;
+    $self->{text} =~ s#<mbp:pagebreak [\s\n]*
+                      #<br style="page-break-after: always" #gix;
+    $self->{text} =~ s#</mbp:pagebreak>##gix;
+    $self->{text} =~ s#</?mbp:nu>##gix;
+    $self->{text} =~ s#</?mbp:section>##gix;
+    $self->{text} =~ s#</?mbp:frameset##gix;
+    $self->{text} =~ s#</?mbp:slave-frame##gix;
 
 
     # More complex alterations will require a HTML tree
     $tree = HTML::TreeBuilder->new();
     $tree->ignore_unknown(0);
+    $tree->no_expand_entities(1);
+    $tree->store_comments(1);
 
-    # If the encoding is UTF-8, we have to decode it before the parse
-    # or the parser will break
-    if($encoding == 65001)
-    {
-        debug(1,"DEBUG: decoding utf8 text");
-        croak($subname,"(): failed to decode UTF-8 text")
-            unless(decode_utf8($$self{text}));
-        $tree->parse($$self{text});
+    # If the encoding is UTF-8, we have to note it specially before
+    # the parse or the parser will break
+    if($encoding == 65001) {
+        $tree->utf8_mode(1);
     }
-    else { $tree->parse($$self{text}); }
+    $tree->parse($self->{text});
     $tree->eof();
 
+    # Update contents of <head>
+    $head_element = $tree->find('head');
+    if(! $head_element) {
+        $self->add_warning("No <head> element found in Mobipocket HTML!");
+    }
+    else {
+        # Fix missing <title>
+        $element = $head_element->find('title');
+        if(! $element) {
+            debug(1,"DEBUG: title element missing, creating as '",
+                  $self->{title},"'");
+            $head_element->push_content(['title',$self->{title}]);
+        }
+        else {
+            if(! $element->as_trimmed_text) {
+                debug(1,"DEBUG: title is empty, setting to '",$self->{title},"'");
+                $element->replace_with($self->{title});
+            }
+        }
+
+        # Set UTF8 content type if necessary
+        if($encoding == 65001) {
+            $head_element->push_content(
+                [ 'meta',
+                {'http-equiv' => 'Content-type'},
+                {'content' => 'text/html;charset=UTF-8' },
+                 ]);
+        }
+    }
+
     # Replace img recindex links with img src links
     debug(2,"DEBUG: converting img recindex attributes");
     @elements = $tree->find('img');
@@ -1386,36 +1431,43 @@ sub fix_html :method   ## no critic (Always unpack @_ first)
     }
 
     debug(2,"DEBUG: converting HTML tree");
-    $$self{text} = $tree->as_HTML;
+    # We don't bother specifying an indent here, as the resulting
+    # formatting is still broken and we have to insert newlines later
+    # anyway.
+    #
+    # TODO: experiment with using Mojo::DOM instead
+    $self->{text} = $tree->as_HTML(undef,'',{});
 
     croak($subname,"(): HTML tree output is empty")
-        unless($$self{text});
+        unless($self->{text});
 
     # Strip embedded nulls
     debug(2,"DEBUG: stripping nulls");
-    $$self{text} =~ s/\0//gx;
+    $self->{text} =~ s/\0//gx;
 
-    # HTML::TreeBuilder will remove all of the newlines, so add some
-    # back in for readability even if tidy isn't called
+    # HTML::TreeBuilder will not insert all appropriate newlines even
+    # specifying indents in as_HTML, so add some back in for
+    # readability even if tidy isn't called
+    #
     # This is unfortunately quite slow.
-    unless($args{nonewlines})
-    {
+    unless($args{nonewlines}) {
         debug(1,"DEBUG: adding newlines");
-        $$self{text} =~ s#<(body|html)> \s* \n?
-                         #<$1>\n#gix;
-        $$self{text} =~ s#\n? <div
-                         #\n<div#gix;
-        $$self{text} =~ s#</div> \s* \n?
-                         #</div>\n#gix;
-        $$self{text} =~ s#</h(\d)> \s* \n?
-                         #</h$1>\n#gix;
-        $$self{text} =~ s#</head> \s* \n?
-                         #</head>\n#gix;
-        $$self{text} =~ s#\n? <(br|p)([\s>])
-                         #\n<$1$2#gix;
-        $$self{text} =~ s#</p>\s*
-                         #</p>\n#gix;
+        $self->{text} =~ s#<(body|html)> \s* \n?
+                          #<$1>\n#gix;
+        $self->{text} =~ s#\n? <div
+                          #\n<div#gix;
+        $self->{text} =~ s#</div> \s* \n?
+                          #</div>\n#gix;
+        $self->{text} =~ s#</h(\d)> \s* \n?
+                          #</h$1>\n#gix;
+        $self->{text} =~ s#</head> \s* \n?
+                          #</head>\n#gix;
+        $self->{text} =~ s#\n? <(br|p)([\s>])
+                          #\n<$1$2#gix;
+        $self->{text} =~ s#</p>\s*
+                          #</p>\n#gix;
     }
+
     return 1;
 }
 
@@ -1465,7 +1517,7 @@ sub fix_html_filepos :method
             # Insert a whole new <a id> here
             debug(2,"DEBUG: filepos ",$pos," points to '<mbp',",
                   " creating new anchor");
-            substr($$self{text},$pos,4,'<a id="' . $pos . '"></a><mbp');
+            substr($$self{text},$pos,4,'<a id="fp' . $pos . '"></a><mbp');
         }
         elsif($atpos =~ /^<(a|p)[ >]/ix)
         {
@@ -1490,7 +1542,7 @@ sub fix_html_filepos :method
             # All other elements
             debug(2,"DEBUG: filepos ",$pos," points to '",$atpos,
                   "', creating new anchor");
-            substr($$self{text},$pos,1,'<a id="' . $pos . '"></a><');
+            substr($$self{text},$pos,1,'<a id="fp' . $pos . '"></a><');
         }
         else
         {
@@ -1620,24 +1672,29 @@ sub find_mobigen
     if($OSNAME eq 'MSWin32')
     {
         @mobigen_guesses = (
+            'kindlegen',
             'mobigen',
+            'C:\Program Files\Mobipocket.com\kindlegen',
             'C:\Program Files\Mobipocket.com\mobigen',
             );
         if($confdir)
         {
             push(@mobigen_guesses,
+                 $confdir . '\kindlegen',
                  $confdir . '\mobigen');
         }
     }
     else
     {
         @mobigen_guesses = (
+            'kindlegen',
             'mobigen',
             'mobigen_linux',
             );
         if($confdir)
         {
             push(@mobigen_guesses,
+                 $confdir . "/kindlegen",
                  $confdir . "/mobigen_linux",
                  $confdir . "/mobigen");
         }
@@ -2598,27 +2655,33 @@ sub record_extradata_size
     my $datalength = length($$dataref);
 
     my $totalsize = 0;
+    my $trailpos = 0;
 
-    foreach my $flagpos (0..15)
+    foreach my $flagbit (reverse 0..15)
     {
-        if($args{extradataflags} & (1 << $flagpos))
+        if($args{extradataflags} & (1 << $flagbit))
         {
             my $bitpos = 0;
+            my $startpos = $trailpos;
             my $traildata;
-            my $trailpos = 0;
             my $trailsize = 0;
             my $byte;
 
             # Bit 0 is the multi-byte character overlap flag, and has
             # a different format from all other flags, where the size
-            # is the first two bits of the last byte of record data.
-            if($flagpos == 0)
+            # is the first two bits of the last unparsed byte of
+            # record data (i.e. the extra data bytes closest to the
+            # actual record text), plus one for the byte containing
+            # the size itself.
+            if($flagbit == 0)
             {
-                $trailsize = ord(substr($$dataref,$datalength-1,1)) & 0x03;
-                debug(3,"DEBUG: ",$trailsize," bytes of trailing data at flagpos ",
-                      $flagpos);
+                $trailpos += 1;
+                $trailsize = ord(substr($$dataref,$datalength-$trailpos,1)) & 0x03;
+                $trailsize += 1; # The above line doesn't include the size byte itself
+                debug(3,"DEBUG: ",$trailsize," bytes of trailing data at flagbit ",
+                      $flagbit,", startpos ",$startpos);
                 $totalsize += $trailsize;
-                next;
+                last;
             }
 
             # For all other bits, the size is a backward-encoded
@@ -2632,8 +2695,8 @@ sub record_extradata_size
             }
             while( !($byte & 0x80) && ($bitpos < 28) && ($trailpos < $datalength) );
             $traildata = substr($$dataref,$datalength-$trailsize,$trailsize);
-            debug(3,"DEBUG: ",$trailsize," bytes of trailing data at flagpos ",
-                  $flagpos);
+            debug(3,"DEBUG: ",$trailsize," bytes of trailing data at flagbit ",
+                  $flagbit,", startpos ",$startpos);
             $totalsize += $trailsize;
         }
     }
@@ -2997,7 +3060,7 @@ sub uncompress_dictionaryhuffman
         {
             croak($subname,"(): invalid HUFF cache offset ",$cacheoffset,
                   " found at bit position ",$bitpos,"!\n");
-            return $text;
+            #return $text;
         }
         $cacheval = $huffref->{cache}->[$cacheoffset];
 #        debug(4,"## cacheval[",$cacheoffset,"]=",$cacheval);
@@ -3007,7 +3070,7 @@ sub uncompress_dictionaryhuffman
         {
             croak($subname,"(): HUFF cache found zero codelength",
                   " at bit position ",$bitpos,"!\n");
-            return $text;
+            #return $text;
         }
 
         $nextbits = min(32,$bitoffset);
@@ -3066,7 +3129,7 @@ sub uncompress_dictionaryhuffman
         {
             croak($subname,"(): \n HUFF entry referenced invalid CDIC ",$cdic,
                   " at bit position ",$bitpos,", depth ",$depth,"!\n");
-            return $text;
+            #return $text;
         }
 
 #        debug(4,"## cdic=",$cdic);
